import argparse
from todocom.tokenize import tokenize
from todocom.utils import get_files, print_todos, save_todos

WHITE = "\033[0m"
YELLOW = "\033[33m"


def main():
    # create a parser object
    parser = argparse.ArgumentParser(description = "Read directories")

    # add argument
    parser.add_argument("dirs",
                        nargs = '*',
                        metavar = "dir",
                        type = str,
                        default = ['./'],
                        help = "Find all files within directories")

    # Urgent priority (activate using -u or --urgent)
    parser.add_argument(
        "-u", "--urgent",
        action="store_true",
        help=("Retrieve 'urgent' TODO comments. "))

    # Soon priority (activate using -s or --soon)
    parser.add_argument(
        "-s", "--soon",
        action="store_true",
        help=("Retrieve 'soon' TODO comments. "))

    # Output file argument
    parser.add_argument(
        "-o", "--output",
        nargs = "?",
        dest = "out",
        const = './todo.txt',
        default = None,
        help = ("Save comments in an external file"))

    # Assigned argument
    parser.add_argument(
        "-a", "--assigned",
        nargs = 1,
        metavar = "assigned_to",
        type = str,
        default = None,
        help = "Find all TODOs that were assigned to a specific user")

    # parse the arguments from standard input
    args = parser.parse_args()

    _ = get_files(args.dirs)
    urgent_todos = []
    reg_todos = []
    soon_todos = []
    todos = []
    for file in _:
        with open(file, "r") as f:
            try:
                for token in tokenize(f.read(),
                                      file,
                                      urgent = args.urgent,
                                      soon = args.soon,
                                      assigned = args.assigned):
                    if args.out:
                        todos.append(token)
                    else:
                        if "urgent" in token.re_type:
                            urgent_todos.append(token)
                        elif "soon" in token.re_type:
                            soon_todos.append(token)
                        else:
                            reg_todos.append(token)
            except UnicodeDecodeError:
                print(f"{YELLOW}**** Warning: Couldn't read file {file} **** \n{WHITE}")

    if args.out:
        # Store results in a txt file
        save_todos(args.out, todos)

    else:
        # Print to terminal
        print_todos(urgent_todos)
        print_todos(soon_todos)
        print_todos(reg_todos)



import os
from pathlib import Path

# Constants
BOLD = "\033[1m"
RED = "\033[31m"
CYAN = "\033[1;36m"
WHITE = "\033[0m"

# Ignore files/dirs lists
ignore_files = ["__main__.py",
                "__init__.py",
                ".DS_Store",
                "pyproject.toml",
                "poetry.lock",
                "README.md",
                ".gitignore",
                "LICENSE"]

ignore_dirs = ["__pycache__",
               "tests",
               ".git",
               ".github",
               "dist"]


# Todo @avivfaraj: test
def format_todos(tokens = [], color = True):
    """
    A generator function that receives list of TODO tokens
    and returns each one as a string with all information.
    'urgent' and 'soon' comments will be printed in RED and
    CYAN respectively if 'color' is True.

    Parameters:
    ------------
    tokens: List[Token]
        List of all TODO tokens

    color: Boolean
        If True -> prints urgent and soon TODO comments in a different
        color to emphasize them.

    Return:
    ------------
    String with information about that TODO comment
    """
    if tokens:
        font_color = WHITE
        end = WHITE
        if color:
            if "urgent" in tokens[0].re_type:
                font_color = BOLD + RED

            elif "soon" in tokens[0].re_type:
                font_color = BOLD + CYAN
            else:
                font_color = WHITE
        else:
            font_color = ""
            end = ""

        for token in tokens:
            if not token.assign:
                yield (f'{token.file} --> pr: {token.re_type}, '
                       f'Line: {token.line}, comment: {font_color} {token.value} {end}')
            else:
                yield (f'{token.file} --> pr: {token.re_type}, assigned: {token.assign}, '
                       f'Line: {token.line}, comment: {font_color} {token.value} {end}')


def print_todos(tokens = []):
    """
    Print all TODOs to terminal
    """
    for i in format_todos(tokens):
        print(i)


def save_todos(file_path, tokens = []):
    """
    Save all TODOs in an external file.
    Default file in current directory is: ./todo.txt
    """
    output_file = Path(file_path).expanduser().resolve()

    # Create new file or override the older file
    f = open(output_file, "w")
    f.close()

    # Append to the new file
    with open(output_file, "a") as file:
        for i in format_todos(tokens, color = False):
            file.write(i + "\n")


def get_files(path_ls = []):
    """
    Generator function that yields all files in a specific directory

    Parameter:
    ----------
    path_ls: List[str]
        List of files/directories to be checked

    Returns:
    ----------
    List of all files to be read and checked for TODO comments
    """
    for path in path_ls:
        if os.path.isfile(path):
            yield path

        if os.path.isdir(path):
            for root, dirs, files in os.walk(path, topdown=True):
                # Remove unwanted directories from list
                [dirs.remove(i) for i in ignore_dirs if i in dirs]

                # Remove unwanted files from list
                [files.remove(i) for i in ignore_files if i in files]

                for file in files:

                    if file not in ignore_files:
                        yield f'{root}/{file}'.replace("//", "/")



from typing import NamedTuple
import re


class Token(NamedTuple):
    re_type: str
    file: str
    value: str
    line: int
    assign: str = None
    # date:


def tokenize(code, file, **kwargs):
    """
    Followed instructions in 're' module documentation and
    modified the code to include TODO regex.

    Types of TODOs:
        1. single_line      ---> should work with most files
        2. multiline        ---> specifically designed for python docstrings
        3. urgent           ---> should work with most files
        4. urgent_multiline ---> specifically designed for python docstrings
        5. soon             ---> should work with most files
        6. soon_multiline   ---> specifically designed for python docstrings
        ** newline          ---> required for counting rows!
    """

    single = '[Tt][Oo0][\\-\\_]?[Dd][Oo0]'
    multi = '\"{3} [Tt][Oo0][\\-\\_]?[Dd][Oo0]'
    assign_to, et_index = None, -1

    # Configurations
    token_specification = [
        ('single_line', '({}:).*'.format(single)),
        ('multiline', '({}:)[\\s\\S]*?(\"\"\")'.format(multi)),
        ('urgent', '({} urgent:).*'.format(single)),
        ('urgent_multiline', '({} urgent:)[\\s\\S]*?(\"\"\")'.format(multi)),
        ('soon', '({} soon:).*'.format(single)),
        ('soon_multiline', '({} soon:)[\\s\\S]*?(\"\"\")'.format(multi)),
        ('assign', '({} @{}).*'.format(single, str(kwargs["assigned"]).lower())),
        ('newline', r'\n')
    ]

    if kwargs["urgent"]:
        tok_regex = '|'.join('(?P<%s>%s)' % pair
                             for pair in token_specification
                             if "urgent" in pair[0] or "newline" in pair[0])

    elif kwargs["soon"]:
        tok_regex = '|'.join('(?P<%s>%s)' % pair
                             for pair in token_specification
                             if "soon" in pair[0] or "newline" in pair[0])

    elif kwargs["assigned"]:
        tok_regex = '|'.join('(?P<%s>%s)' % pair
                             for pair in token_specification
                             if "assign" in pair[0] or "newline" in pair[0])

    else:
        tok_regex = '|'.join('(?P<%s>%s)' % pair
                             for pair in token_specification
                             if "assign" not in pair[0])

    line_num = 1
    multi_line_num = None
    end_last_token = -1

    # iteration over tokens found in code
    for mo in re.finditer(tok_regex, code):

        # Ensure last token was multi-line
        if multi_line_num:

            # Must add the number of rows in the comment
            # Or else the next tokens will be shifted!
            line_num += multi_line_num
            multi_line_num = None

        # Type of token (TODO) found
        kind = mo.lastgroup

        # Ensure comment is an actual TODO
        if kind != "newline":

            # Store TODO comment
            value = mo.group()

            # Clean comment and find its starting index
            value = re.sub("\n", " ", value.strip())
            value = re.sub(" +", " ", value.strip())

            index = re.search(":", value).span()[0]

            if kind == "assign":
                et_index = re.search("@", value).span()[0]

            multi_line_num = 1

            # Ensure ":" was found
            if index:
                index = re.search(":", value).span()[0]

            # Ensure TODO is multiline
            if 'multiline' in kind:

                # Calculate number of lines to add
                multi_line_num = len(re.findall("\n", mo.group())) + 1

                # Store the comment without the trailing double quotes (""")
                value = value[index + 1: -3].strip()

            else:
                if kind != "assign":
                    assign_to = None
                else:
                    assign_to = value[et_index + 1: index].strip()

                # Comment is single line --> store it all
                value = value[index + 1:].strip()

            # Store the last char in the token
            # Required to prevent "\n" from being considered twice
            end_last_token = mo.end()

            yield Token(kind, file, value, line_num, assign_to)

        # Ensure token is a newline
        if kind == 'newline':

            # Ensure newline is not used twice
            if mo.start() != end_last_token:
                line_num += 1
            else:
                end_last_token = -1



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['todocom']

package_data = \
{'': ['*']}

entry_points = \
{'console_scripts': ['todo = todocom.cli:main']}

setup_kwargs = {
    'name': 'todocom',
    'version': '0.2.7',
    'description': 'CLI to retrieve a list of all TODO comments in the code',
    'long_description': '# todocom (Todo Comments)\nCLI program that retrieves all TODO comments from file(s) and prints them in terminal/shell. It was created in order to automatically update a list of TODO tasks by simply adding "TODO:" comments in the code. It also enables prioritization of tasks by using "TODO soon:" or "TODO urgent". \nTo create the TODO list, simply open terminal and run the following command:\n```\ntodo [folder/file]\n```\n\nThis command will print out all TODO comments that were found in the code, sorted by their prioritization: urgent, soon and regular. \n_Urgent_ tasks will be printed in RED, _soon_ in CYAN and _regular_ comments in WHITE to make it easier to read. There is also an option to filter comments by their priotization:\n```\n# Prints urgent TODOs\ntodo -u [folder/file]\n```\nOr:\n```\n# Prints soon TODOs\ntodo -s [folder/file]\n```\n\nFinally, there is an option to save the list in a text file (stores as regular text without colors):\n```\n# Store results in a txt file\ntodo -o [path/to/sample.txt] [folder/file]\n```\n\n\n## Setup\n```\npip install todocom\n```',
    'author': 'avivfaraj',
    'author_email': 'avivfaraj4@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/avivfaraj/todocom',
    'packages': packages,
    'package_data': package_data,
    'entry_points': entry_points,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)
